"""
launch_server.py

access igeLauncher apprication server
"""

import requests
import json
import hashlib
import os.path
import glob

#LAUNCH_SERVER_URL = 'http://127.0.0.1:8080'
LAUNCH_SERVER_URL = 'https://igelauncher.appspot.com'

def mkdir(dir,name):
    """
    Make a directory on the server

    :param dir:  parent directory name
    :param name: directory to create
    :return: 0 if success , not 0 if error
    """
    url = LAUNCH_SERVER_URL + '/mkdir'
    param = {'directory':dir,'name':name}
    try:
        response = requests.post(url, param)
    except:
        print('mkdir ; network error')
        return -1

    if response.status_code != 200:
        print('mkdir ; network error')
        if response.text is not None:
            print(response.text)
        return -1
    return 0

def ls(dir):
    """
    Lists directory contents of files and directories on the server

    :param dir: directory path
    :return: files, err
              files : {filename : {type:filetype, hash:hashvalue, sizse:filesize}}
               err: not 0 if error occurred
    """
    url = LAUNCH_SERVER_URL + '/ls'
    param = {'directory': dir}
    try:
        response = requests.post(url, param)
    except:
        return None, -1
    if response.status_code == 200:
        list = json.loads(response.text)
        return list,0
    else:
        print('Network error! Status Code: {:d}'.format(response.status_code))
        if response.text is not None:
            print(response.text)
    return None, -1

def download_files(src,dst):
    """
    Download all files in the specified directory on the server

    :param src: server directory path
    :param dst: local direcgtory path to save files
    :return: not 0 if error occurred
    """
    list,err = ls(src)
    if list is None: return err

    url = LAUNCH_SERVER_URL+'/download_request'

    for key, val in list.items():
        localHash = ""
        path = dst+'/'+key
        if os.path.exists(path):
            with open(path, 'rb') as f:
                fileDataBinary = f.read()
                if len(fileDataBinary) != 0:
                    hash = hashlib.md5()
                    hash.update(fileDataBinary)
                    localHash = hash.hexdigest()

        if localHash != val['hash']:
            param = {'directory':src,'file':key}

            try:
                response = requests.post(url, param)
                dlurl = response.text
                response = requests.get(dlurl)
            except:
                print('download {} was failed'.format(key))
                return -1

            if response.status_code == 200:
                with open(path, 'wb') as f:
                    f.write(response.content)
                print('downloaded {}'.format(key))
            else:
                print('Network error! Status Code: {:d}'.format(response.status_code))
                if response.text is not None:
                    print(response.text)
        else:
            print('download {} was skipped because no change'.format(key))
    return 0

def upload_files(src,dst):
    url = LAUNCH_SERVER_URL + '/ls'
    param = {'directory': dst}
    response = requests.post(url, param)

    serverFiles = {}
    if response.status_code == 200:
        serverFiles = json.loads(response.text)

    localFiles = []
    list = glob.glob(os.path.join(src, '*'))
    for f in list:
        if os.path.isfile(f):
            localFiles.append(f)
    for path in localFiles:
        fileDataBinary = ''
        with open(path, 'rb') as f:
            fileDataBinary = f.read()

        filesize = len(fileDataBinary)
        if filesize == 0: continue

        hash = hashlib.md5()
        hash.update(fileDataBinary)
        hashcode = hash.hexdigest()

        f = os.path.basename(path)
        if f not in serverFiles or serverFiles[f]['hash'] != hashcode:

            try:
                if filesize > 32*1024*1024:
                    url = LAUNCH_SERVER_URL + '/upload_start'
                    param = {'directory': dst, 'filename': os.path.basename(path), 'filesize':filesize, 'hash':hashcode}
                    response = requests.post(url, param)
                    url = response.text
                    response = requests.post(url, data=fileDataBinary)
                    url = LAUNCH_SERVER_URL + '/upload_finish'
                    response = requests.post(url, param)
                else:
                    url = LAUNCH_SERVER_URL + '/upload'
                    XLSX_MIMETYPE = 'application/vnd.openxmlformats-officedocument.spreadsheetml.sheet'
                    files = {'file': (f, fileDataBinary, XLSX_MIMETYPE)}
                    param = {'directory': dst, 'hash': hashcode}
                    response = requests.post(url, param, files=files)
            except:
                print('upload {} was failed'.format(f))
                return -1

            if response.status_code == 200:
                print(path + " uploaded.")
            else:
                print('Network error! Status Code {:d}\n'.format(response.status_code))
                if response.text is not None:
                    print(response.text)
        else:
            print(path + " is not changed.")

    return 0

def makeProjectDir(user, app, version, platform):
    root = '/'
    if mkdir(root, user) == 0:
        root += user
        if mkdir(root, app) == 0:
            root += '/'
            root += app
            if mkdir(root, version) == 0:
                root += '/'
                root += version
                if mkdir(root, platform)==0:
                    return True
    return False



"""
graphicsHelper.py

Functions to help build graphics data
"""

import igeCore as core
import igeVmath as vmath
import math

_uvs = (0.0, 1.0, 1.0, 1.0, 0.0, 0.0, 1.0, 0.0)
_tris = (0, 2, 1, 1, 2, 3)
_pivotoffset = (1,-1, 0,-1, -1,-1, 1,0, 0,0, -1,0, 1,1, 0,1, -1,1)

def textFigure(word, fontpath, fontsize,color=(1.0, 1.0, 1.0, 1.0), pivot=4, scale = 1.0):
    '''
    Create Visible text

    Parameters
    ----------
        word : string
            text
        fontpath : string
            font data file path
        fontsize : string
            font size
        color : (float,float,float,float)
            text color
        pivot : int
            center point of polygon
            0       1       2
             +-------+-------+
             |       |       |
            3|      4|      5|
             +-------+-------+
             |       |       |
            6|      7|      8|
             +-------+-------+
    '''

    w,h = core.calcFontPixelSize(word, fontpath, fontsize)
    tex = core.texture(core.unique("text"),w,h, format=core.GL_RED)
    tex.setText(word, fontpath,fontsize)

    gen = core.shaderGenerator()
    gen.setColorTexture(True)
    gen.setBoneCondition(1, 1)
    gen.discardColorMapRGB()

    hw = w/2 * scale
    hh = h/2 * scale
    px = _pivotoffset[pivot*2+0] * hw;
    py = _pivotoffset[pivot*2+1] * hh;

    points = ((-hw+px,hh+py,0.0), (hw+px,hh+py,0.0), (-hw+px,-hh+py,0.0), (hw+px,-hh+py,0.0))

    efig = createMesh(points, _tris, None, _uvs, gen)

    efig.setMaterialParam("mate", "DiffuseColor", color);
    efig.setMaterialParamTexture("mate", "ColorSampler", tex,
                                    wrap_s=core.SAMPLERSTATE_BORDER,wrap_t=core.SAMPLERSTATE_BORDER,
                                    minfilter=core.SAMPLERSTATE_NEAREST, magfilter=core.SAMPLERSTATE_LINEAR)
    efig.setMaterialRenderState("mate", "blend_enable", True)

    return efig

def createSprite(width:float=100, height:float=100, texture=None, uv_left_top:tuple=(0,0), uv_right_bottom:tuple=(1,1), normal=None, pivot=4, shader=None):
    """
    Create Visible Rectangle

	Parameters
	----------
        width : float (optional)
            sprite width
        height: float (optional)
            sprite height
        texture: string (optional)
            texture file name
        uv_left_top: tuple (optional)
            texture uv value of left top cornar
        uv_right_bottom: tuple (optional)
            texture uv value of right bottom cornar
        normal: pyvmath.vec3 (optional)
            specify rectangle's nomal vector
        pivot : int
            center point of polygon
        shader : igeCore.shader
            shader object
	Returns
	-------
        editableFigure
    """
    hw = width/2
    hh = height/2
    px = _pivotoffset[pivot*2+0] * hw;
    py = _pivotoffset[pivot*2+1] * hh;
    points = ((-hw+px,hh+py,0.0), (hw+px,hh+py,0.0), (-hw+px,-hh+py,0.0), (hw+px,-hh+py,0.0))

    if normal is not None:
        newpoints = []
        nom0 = (0, 0, 1)
        mat = vmath.mat33(vmath.quat_rotation(nom0, normal))
        for p in points:
            newpoints.append(mat * p)
        points = newpoints

    # uvs = (uv_left_top[0], uv_right_bottom[1],
    #        uv_right_bottom[0], uv_right_bottom[1],
    #        uv_left_top[0], uv_left_top[1],
    #        uv_right_bottom[1], uv_left_top[1])
    uvs = ( uv_left_top[0], uv_right_bottom[1],
            uv_right_bottom[0], uv_right_bottom[1],
            uv_left_top[0], uv_left_top[1],
            uv_right_bottom[0], uv_left_top[1])

    return createMesh(points, _tris, texture, uvs, shader)

def createMesh(points, tris, texture=None, uvs = None, shader = None, normals = None):
    """
    Create a polygon mesh by specifying vertex coordinates and triangle index

	Parameters
	----------
        points: tuple or list
            list or tuple of points
        tris: tuple or list
            list or tuple of triangle indices
        texture: string (optional)
            file path of texture
        uvs: list or tuple (optional)
        shader : igeCore.shader
            shader object
        normals : list or tuple (optional)
            list or tuple of  vertex normal

	Returns
	-------
        editableFigure
   
    """
    if shader is None:
        shader = core.shaderGenerator()
        if texture != None:
            shader.setColorTexture(True)
        shader.setBoneCondition(1, 1)

    efig = core.editableFigure("sprite", True)
    efig.addMaterial("mate", shader)
    efig.addMesh("mesh", "mate");

    efig.setVertexElements("mesh", core.ATTRIBUTE_ID_POSITION, points)
    if uvs:
        efig.setVertexElements("mesh", core.ATTRIBUTE_ID_UV0,uvs)

    if  normals:
        efig.setVertexElements("mesh", core.ATTRIBUTE_ID_NORMAL,normals)

    efig.setTriangles("mesh", tris);
    efig.addJoint("joint");
    efig.setMaterialParam("mate", "DiffuseColor", (1.0, 1.0, 1.0, 1.0));
    #efig.setMaterialRenderState("mate", "cull_face_enable", False)

    if texture != None:
        efig.setMaterialParamTexture("mate", "ColorSampler", texture,
                                     wrap_s=core.SAMPLERSTATE_BORDER,wrap_t=core.SAMPLERSTATE_BORDER,
                                     minfilter=core.SAMPLERSTATE_LINEAR, magfilter=core.SAMPLERSTATE_LINEAR)
        efig.setMaterialRenderState("mate", "blend_enable", True)

    return efig

def __tranform(normal, poss, noms):
	nom0 = (0, 0, 1)

	newpoints = []
	mat = vmath.mat33(vmath.quat_rotation(nom0, normal))
	for p in poss:
		newpoints.append(mat * p)
	poss = newpoints
	newnoms = []
	for n in noms:
		newnoms.append(mat * n)
	noms = newnoms

	return poss, noms

###################################################################################################
plane_triangles=(0, 2, 1, 1, 2, 3)
def makePlane(width:float, height:float, uv_left_top:tuple=(0,0), uv_right_bottom:tuple=(1,1), normal=None, jointIndex=0):
    w = width / 2
    h = height / 2
    poss = (vmath.vec3(-w, h, 0.0), vmath.vec3(w, h, 0.0), vmath.vec3(-w, -h, 0.0), vmath.vec3(w, -h, 0.0))

    uvs = (vmath.vec2(uv_left_top[0], uv_right_bottom[1]),
            vmath.vec2(uv_right_bottom[0], uv_right_bottom[1]),
            vmath.vec2(uv_left_top[0], uv_left_top[1]),
            vmath.vec2(uv_right_bottom[1], uv_left_top[1]))

    nom = vmath.vec3(0,0,1)
    noms = (nom,)*4

    if normal is not None:
        poss, noms = __tranform(normal, poss, noms)

    idxs = ((jointIndex,0,0,0),)*len(poss)

    return poss, noms, uvs, idxs, plane_triangles

###################################################################################################
cubeN = ((-1.0, 0.0, 0.0), (0.0, 1.0, 0.0), (1.0, 0.0, 0.0),
		 (0.0, -1.0, 0.0), (0.0, 0.0, 1.0), (0.0, 0.0, -1.0))
cubeF = ((0, 1, 5, 4), (4, 5, 6, 7), (7, 6, 2, 3),
		 ( 1, 0, 3, 2 ), ( 1, 2, 6, 5 ), ( 0, 4, 7, 3 ))
cubeV = ((-.5, -.5, -.5), (-.5, -.5,  .5), ( .5, -.5,  .5), ( .5, -.5, -.5),	# Lower tier (lower in y)
		 (-.5, .5, -.5), (-.5, .5,  .5), ( .5, .5, .5), ( .5, .5, -.5)) 		# Upper tier

cubeT = ((0.0, 0.0), (0.0, 1.0), (1.0, 1.0), (1.0, 0.0))

def makeBox(width:float, height:float, depth:float, jointIndex=0):
    poss = []
    noms = []
    uvs = []
    tris = []

    nvrt = 0
    for i in range(6):
        for j in range(4):
            poss.append(vmath.vec3(cubeV[cubeF[i][j]][0] * width, cubeV[cubeF[i][j]][1] * height, cubeV[cubeF[i][j]][2] * depth))
            noms.append(vmath.vec3(cubeN[i][0], cubeN[i][1], cubeN[i][2]))
            uvs.append(vmath.vec2(cubeT[j][0], cubeT[j][1]))
        tris.append(nvrt)
        tris.append(nvrt + 1)
        tris.append(nvrt + 2)
        tris.append(nvrt + 2)
        tris.append(nvrt + 3)
        tris.append(nvrt)
        nvrt += 4

    idxs = ((jointIndex,0,0,0),)*len(poss)

    return poss, noms, uvs, idxs, tris

###################################################################################################
def makeBoxFromAABB(min , max, jointIndex=0):
    cubeV2 = ((min.x, min.y, min.z), (min.x, min.y,  max.z), (max.x, min.y, max.z), (max.x, min.y, min.z),
                (min.x, max.y, min.z), (min.x, max.y,  max.z), (max.x, max.y, max.z), (max.x, max.y, min.z))
    poss = []
    noms = []
    uvs = []
    tris = []

    nvrt = 0
    for i in range(6):
        for j in range(4):
            poss.append((cubeV2[cubeF[i][j]][0], cubeV2[cubeF[i][j]][1], cubeV2[cubeF[i][j]][2]))
            noms.append((cubeN[i][0], cubeN[i][1], cubeN[i][2]))
            uvs.append((cubeT[j][0], cubeT[j][1]))

        tris.append(nvrt)
        tris.append(nvrt + 1)
        tris.append(nvrt + 2)
        tris.append(nvrt + 2)
        tris.append(nvrt + 3)
        tris.append(nvrt)
        nvrt += 4

    idxs = ((jointIndex,0,0,0),)*len(poss)

    return poss, noms, uvs, idxs, tris

###################################################################################################
def makeCylinder(radius1:float, radius2:float, length:float, slices:int, stacks:int, normal=None, jointIndex=0):

    # Sin/Cos caches
    sinI = []
    cosI = []
    for i in range(slices):
        angle = 2.0 * math.pi * i / slices
        sinI.append(math.sin(angle))
        cosI.append(math.cos(angle))

    # Compute side normal angle
    deltaRadius = radius2 - radius1
    sideLength = math.sqrt( deltaRadius * deltaRadius + length * length )

    normalXY = 1.0
    if sideLength > 0.00001:
        normalXY =  length / sideLength

    normalZ = 0.0
    if sideLength > 0.00001:
        normalZ = deltaRadius / sideLength

    # Base cap (uSlices + 1)
    fZ = length * -0.5
    radius = radius1

    poss = []
    noms = []
    uvs = []
    tris = []

    poss.append((0.0, 0.0, fZ))
    noms.append((0.0, 0.0, -1.0))
    uvs.append((0.0, 0.0))

    for i in range(slices):
        poss.append((radius * sinI[i], radius * cosI[i], fZ))
        noms.append((0.0, 0.0, -1.0))
        uvs.append((0.0, 0.0))

    # Stacks ((uStacks + 1)*uSlices)
    for j in range(stacks+1):
        f = j / stacks
        fZ = length * ( f - 0.5 )
        radius = radius1 + f * deltaRadius

        for i in range(slices):
            poss.append((radius * sinI[i], radius * cosI[i], fZ))
            noms.append((normalXY * sinI[i], normalXY * cosI[i], normalZ))
            uvs.append((0.0, 0.0))

    # Top cap (uSlices + 1)
    fZ = length * 0.5
    radius = radius2

    for i in range(slices):
        poss.append((radius * sinI[i], radius * cosI[i], fZ))
        noms.append((0.0, 0.0, 1.0))
        uvs.append((0.0, 0.0))
    poss.append((0.0, 0.0, fZ))
    noms.append((0.0, 0.0, 1.0))
    uvs.append((0.0, 0.0))

    if normal is not None:
        poss, noms = __tranform(normal, poss, noms)


    # Generate indices

    # Z+ pole (uSlices)
    rowA = 0
    rowB = 1

    for i in range(slices - 1):
        tris.append( rowA )
        tris.append( rowB + i )
        tris.append( rowB + i + 1 )

    tris.append( rowA )
    tris.append( rowB + slices - 1 )
    tris.append( rowB )

    # Interior stacks (uStacks * uSlices * 2)
    for j in range(stacks):
        rowA = 1 + ( j + 1 ) * slices
        rowB = rowA + slices
        for i in range(slices - 1):
            tris.append( rowA + i )
            tris.append( rowB + i )
            tris.append( rowA + i + 1 )
            tris.append( rowA + i + 1 )
            tris.append( rowB + i )
            tris.append( rowB + i + 1 )

        tris.append( rowA + slices - 1 )
        tris.append( rowB + slices - 1 )
        tris.append( rowA )

        tris.append( rowA )
        tris.append( rowB + slices - 1 )
        tris.append( rowB )

    # Z- pole (uSlices)
    rowA = 1 + ( stacks + 2 ) * slices
    rowB = rowA + slices

    for i in range(slices - 1):
        tris.append( rowA + i )
        tris.append( rowB )
        tris.append( rowA + i + 1 )

    tris.append( rowA + slices - 1 )
    tris.append( rowB )
    tris.append( rowA )

    idxs = ((jointIndex,0,0,0),) * len(poss)

    return poss, noms, uvs, idxs, tris

###################################################################################################
def makeSphere(radius:float, slices:int, stacks:int, jointIndex=0 ):

    # Sin/Cos caches
    sinI = []
    cosI = []
    sinJ = []
    cosJ = []

    for i in range(slices):
        angle = 2.0 * math.pi * i / slices
        sinI.append(math.sin(angle))
        cosI.append(math.cos(angle))

    for j in range(stacks):
        angle = math.pi * j / stacks
        sinJ.append(math.sin(angle))
        cosJ.append(math.cos(angle))

    poss = []
    noms = []
    uvs = []
    tris = []

    # +Z pole
    poss.append(vmath.vec3(0.0, 0.0, radius))
    noms.append(vmath.vec3(0.0, 0.0, 1.0))
    uvs.append(vmath.vec2(0.0, 0.0))

    # Stacks
    for j in range(stacks):
        for i in range(slices):
            norm = vmath.vec3(sinI[i] * sinJ[j], cosI[i] * sinJ[j], cosJ[j])
            poss.append(vmath.vec3(norm.x * radius, norm.y * radius, norm.z * radius))
            noms.append(norm)
            uvs.append(vmath.vec2(0.0, 0.0))

    # Z- pole
    poss.append(vmath.vec3(0.0, 0.0, -radius))
    noms.append(vmath.vec3(0.0, 0.0, -1.0))
    uvs.append(vmath.vec2(0.0, 0.0))

    # Generate indices
    # Z+ pole
    rowA = 0
    rowB = 1

    # for i in range(slices-1):
    #     tris.append( rowA )
    #     tris.append( rowB + i + 1 )
    #     tris.append( rowB + i )
    # tris.append( rowA )
    # tris.append( rowB )
    # tris.append( rowB + slices-1 )

    # Interior stacks
    for j in range(1,stacks):
        rowA = 1 + (j - 1) * slices
        rowB = rowA + slices
        for i in range(slices - 1):
            tris.append(rowA + i)
            tris.append(rowA + i + 1)
            tris.append(rowB + i)

            tris.append(rowA + i + 1)
            tris.append(rowB + i + 1)
            tris.append(rowB + i)

        tris.append(rowA + slices - 1)
        tris.append(rowA)
        tris.append(rowB + slices - 1)

        tris.append(rowA)
        tris.append(rowB)
        tris.append(rowB + slices - 1)

    #Z- pole
    rowA = 1 + (stacks - 1) * slices
    rowB = rowA + slices

    for i in range(slices - 1):
        tris.append(rowA + i)
        tris.append(rowA + i + 1)
        tris.append(rowB)

    tris.append(rowA + slices - 1)
    tris.append(rowA)
    tris.append(rowB)

    idxs = ((jointIndex,0,0,0),) * len(poss)

    return poss, noms, uvs, idxs, tris

###################################################################################################
def makeTorus(innerRadius:float, outerRadius:float, sides:int, rings:int, normal=None, jointIndex=0):

    poss = []
    noms = []
    uvs = []
    tris = []

    # Compute the vertices
    for i in range(rings):
        theta = i * 2.0 * math.pi / rings
        st = math.sin(theta)
        ct = math.cos(theta)
        for j in range(sides):
            phi = j * 2.0 * math.pi / sides
            sp = math.sin(phi)
            cp = math.cos(phi)
            poss.append((ct * ( outerRadius + innerRadius * cp ), -st * ( outerRadius + innerRadius * cp ), sp * innerRadius))
            noms.append((ct * cp, -st * cp, sp))
            uvs.append((0.0, 0.0))

    if normal is not None:
    	poss, noms = __tranform(normal, poss, noms)

    for i in range(rings - 1):
        for j in range(sides - 1):
            # Tri 1 (Top-Left tri, CCW)
            tris.append( i * sides + j )
            tris.append( i * sides + j + 1 )
            tris.append( ( i + 1 ) * sides + j )

            # Tri 2 (Bottom-Right tri, CCW)
            tris.append( ( i + 1 ) * sides + j )
            tris.append( i * sides + j + 1 )
            tris.append( ( i + 1 ) * sides + j + 1 )

        j = sides - 1
        # Tri 1 (Top-Left tri, CCW)
        tris.append( i * sides + j )
        tris.append( i * sides )
        tris.append( ( i + 1 ) * sides + j )

        # Tri 2 (Bottom-Right tri, CCW)
        tris.append( ( i + 1 ) * sides + j )
        tris.append( i * sides + 0 )
        tris.append( ( i + 1 ) * sides + 0 )

    i = rings-1
    #join the two ends of the tube
    for j in range(sides - 1):
        # Tri 1 (Top-Left tri, CCW)
        tris.append( i * sides + j )
        tris.append( i * sides + j + 1 )
        tris.append( j )

        # Tri 2 (Bottom-Right tri, CCW)
        tris.append( j )
        tris.append( i * sides + j + 1 )
        tris.append( j + 1 )

    j = sides-1
    # Tri 1 (Top-Left tri, CCW)
    tris.append( i * sides + j )
    tris.append( i * sides )
    tris.append( j )

    # Tri 2 (Bottom-Right tri, CCW)
    tris.append( j )
    tris.append( i * sides )
    tris.append( 0 )

    idxs = ((jointIndex,0,0,0),) * len(poss)

    return poss, noms, uvs, idxs, tris




"""
indi game engine offline develop tools

convert 3d assets
convert images to platform dipend format
make ige package
upload package to devserver

"""

import py_compile
import shutil
import glob
import os.path
import igeCore as core
from igeCore import apputil
from igeCore.devtool._igeTools import *
from igeCore import apputil
from igeCore.apputil import launch_server
import hashlib
import tokenize

def isHiddenFolder(path):
    if path.find('/.') != -1 or path.find('\\.') != -1 :
        return True
    return False

def findFiles(path, exts = ['.*']):
    list = []
    for root, dirs, files in os.walk(path):
        if isHiddenFolder(root): continue
        for fname in files:
            name, ext = os.path.splitext(fname)
            for e in exts:
                if e == '.*' or e == ext:
                    list.append(os.path.join(root, fname))
                    break
    return list

def findImageFiles(path):
    list = []
    for root, dirs, files in os.walk(path):
        if isHiddenFolder(root): continue
        for fname in files:
            name, ext = os.path.splitext(fname)
            extl = ext.lower()
            if extl == '.png' or extl == '.dds' or extl == '.tga' or extl == '.psd' or extl == '.jpg':
                list.append(os.path.join(root, fname))
    return list

def find3DSceneFiles(path):
    dict = {}
    for root, dirs, files in os.walk(path):
        if root.find('\.') != -1: continue
        folder = os.path.basename(root)

        folderRule = False
        for f in files:
            name, ext = os.path.splitext(f)
            if name == folder:
                folderRule = True
                baseFile = f
                break

        if folderRule:
            daes = [os.path.join(root, baseFile)]
            for f2 in files:
                f2_name, f2_ext = os.path.splitext(f2)
                if (f2_ext.lower() == '.dae' or f2_ext.lower() == '.fbx') and f2_name != folder:
                    daes.append(os.path.join(root, f2))
            dict[folder] = daes
        else:
            for f2 in files:
                f2_name, f2_ext = os.path.splitext(f2)
                if f2_ext.lower() == '.dae' or f2_ext.lower() == '.fbx':
                    dict[f2_name] = [os.path.join(root, f2)]
    return dict

def createDirectoryTree(filelist, root):
    imageDirectory = dict()
    for img in filelist:
        while True:
            img, file = os.path.split(img)
            if img not in imageDirectory:
                imageDirectory[img] = {file: 0};
            elif file not in imageDirectory[img]:
                imageDirectory[img][file] = 0
            if img == root: break
    return imageDirectory

def findFileFromDiorectoryTree(imageDirectory, file, dir):
    files = imageDirectory.get(dir)
    if files is None: return None
    if file in files:
        return os.path.join(dir, file)
    for d in files:
        _, ext = os.path.splitext(d)
        if ext == '':
            rv = findFileFromDiorectoryTree(imageDirectory, file, os.path.join(dir, d))
            if rv: return rv
    return None

def removeRoot(path, root):
    newpath = path.replace(root, '', 1)
    if len(newpath) >0:
        if newpath[0] == '/' or newpath[0] == '\\':
            newpath = newpath[1:]
    else:
        newpath = ""
    return newpath

def replaceExt(file, ext):
    name, extold = os.path.splitext(file)
    return name + ext

def readAndCalcHash(srcfile, dstfile, hashfile):
    hashcode = '00000000000000000000000000000000'
    newhash  = 'ffffffffffffffffffffffffffffffff'
    if os.path.exists(dstfile):
        try:
            with open(hashfile, 'r') as f:
                hashcode = f.read()
        except : None

    with open(srcfile, 'rb') as f:
        fileDataBinary = f.read()
        if len(fileDataBinary) != 0:
            hash = hashlib.md5()
            hash.update(fileDataBinary)
            newhash = hash.hexdigest()
    return hashcode,newhash

def saveHash(hashfile, newhash):
    if newhash != 'ffffffffffffffffffffffffffffffff':
        apputil.makeDirectories(hashfile)
        with open(hashfile, 'w') as f:
            f.write(newhash)

def readFigureConf(path):
    mode = 0
    label = ''
    keyvalue = {}
    with open(path, 'rb') as f:
        for token in tokenize.tokenize(f.readline):
            if token.type is tokenize.ENDMARKER:
                break
            elif token.type is tokenize.NAME:
                if mode == 0:
                    label = token.string
                else:
                    keyvalue[label] = token.string
                    mode = 0
            elif token.type is tokenize.OP:
                mode = 1
            elif token.type is tokenize.NUMBER:
                if mode == 1:
                    try:
                        keyvalue[label] = int(token.string)
                    except ValueError:
                        keyvalue[label] = float(token.string)
                    mode = 0
    return keyvalue

def createFigureConfTree(path, unit=1.0):
    path = os.path.abspath(path)
    filelist = []
    for root, dirs, files in os.walk(path):
        if root.find('\.') != -1: continue
        for fname in files:
            if fname == 'figure.conf':
                filelist.append(os.path.join(root, fname))
    figureConfTree = dict()
    for filepath in filelist:
        conf = readFigureConf(filepath)
        dir, file = os.path.split(filepath)
        figureConfTree[dir] = conf

    for dir, val in figureConfTree.items():
        while True:
            dir, _ = os.path.split(dir)
            if dir == os.path.abspath('.').split(os.path.sep)[0]+os.path.sep:
                break
            parent = figureConfTree.get(dir)
            if parent is not None:
                for parentkey, parentval in parent.items():
                    if val.get(parentkey) is None:
                        val[parentkey] = parentval

    if len(figureConfTree) == 0:
        figureConfTree[path] = {}

    for conf in figureConfTree.values():
        if conf.get('BASE_SCALE') is None:
           conf['BASE_SCALE'] = unit
        #if conf.get('OPTIMIZE_MESH') is None:
        #   conf['OPTIMIZE_MESH'] = True

    return figureConfTree

def findFigureConf(dir: str, tree: dict):
    dir = os.path.abspath(dir)
    while True:
        dir, _ = os.path.split(dir)
        if dir == os.path.abspath('.').split(os.path.sep)[0]+os.path.sep:
            break
        conf = tree.get(dir)
        if conf is not None:
            return conf
    return dict()

def load3DScene(tmpfig, files, conf):
    loadModel(files[0], tmpfig, conf)
    for i in range(1, len(files)):
        loadAnimation(files[i], tmpfig, conf)

def replaceTexturePath(tmpfig, imageDirectory, startPath, endPath, allTextures, subFolder):
    texs = tmpfig.getTextureSource()
    for tex in texs:
        filename = os.path.basename(tex['path'])
        current = startPath
        inputimage = None
        while True:
            inputimage = findFileFromDiorectoryTree(imageDirectory, filename, current)
            if inputimage is not None: break
            if current == endPath: break
            current, _ = os.path.split(current)
        if inputimage is None:
            print('file not found ' + tex['path'])
        else:
            allTextures[inputimage] = tex;
            inputimage = removeRoot(inputimage, endPath);
            inputimage, ext = os.path.splitext(inputimage)
            inputimage = os.path.join(subFolder, inputimage)
            inputimage = inputimage.replace('\\', '/')
            newTex = {'path': inputimage, 'normal': tex['normal'], 'wrap': tex['wrap']}
            tmpfig.replaceTextureSource(tex, newTex)

def setMaterialAlpha(tmpfig, imageDirectory, startPath, endPath):
    texs = tmpfig.getTextureSource()
    for tex in texs:
        filename = os.path.basename(tex['path'])
        current = startPath
        inputimage = None
        while True:
            inputimage = findFileFromDiorectoryTree(imageDirectory, filename, current)
            if inputimage is not None: break
            if current == endPath: break
            current, _ = os.path.split(current)
        if inputimage is not None:
            if isAlpha(inputimage):
                tmpfig.enableAlphaModeByTexture(tex['path'])

def getHashPath(root, source, platform):
    abspath = os.path.abspath(source)
    if abspath[0]=='/' or abspath[0]=='\\':
       abspath = abspath[1:]
    elif abspath[2]=='\\':
        abspath = abspath[3:]
    return os.path.join(os.path.join(root, apputil.targetFolderName(platform)), abspath)

def convertImages(allTextures, sourceDir, destDir, platform):
    print('----------------------')
    print('Convert image files')
    print('----------------------')
    for key, val in allTextures.items():
        outfile = os.path.normpath(key.replace(sourceDir, destDir, 1))
        outfile = replaceExt(outfile, '.pyxi')
        hashfile = getHashPath('.tmp/hash', key, platform)
        hashcode, newhash = readAndCalcHash(key, outfile, hashfile)
        if hashcode != newhash:
            print("convert {} for {}".format(key,apputil.targetFolderName(platform)))
            apputil.makeDirectories(outfile)
            convertTextureToPlatform(key, outfile, platform, val['normal'], val['wrap'])
            saveHash(hashfile, newhash)
        else:
            print("skip {} ".format(key))

def copyFiles(src, dst, exts=['.*']):
    for root, dirs, files in os.walk(src):
        if isHiddenFolder(root): continue
        for fname in files:
            name, ext = os.path.splitext(fname)
            for e in exts:
                if e == '.*' or ext == e:
                    srcPath = os.path.join(root, fname)
                    dstPath = os.path.join(dst, removeRoot(srcPath, src))
                    apputil.makeDirectories(dstPath)
                    shutil.copy2(srcPath, dstPath)
                    print("copy {}".format(dstPath))
                    break

def compilePrograms(src, dst):
    list = findFiles(src, ['.py'])
    for infile in list:
        outfile = os.path.normpath(infile.replace(src, dst, 1))
        apputil.makeDirectories(outfile)
        if os.path.basename(infile) == 'root.py':
            shutil.copy(infile, outfile)
        else:
            outfile = replaceExt(outfile, '.pyc')
            apputil.makeDirectories(outfile)
            print('compile : '+infile)
            py_compile.compile(infile, outfile)

def hashCheck(files, outfile, platform):
    changed = False
    for file in files:
        hashfile = getHashPath('.tmp/hash', file, platform)
        hashcode, newhash = readAndCalcHash(file, outfile, hashfile)
        if hashcode != newhash:
            changed = True
            saveHash(hashfile, newhash)
    return changed

def buildAppPackage(src, dest):
    apputil.makeDirectories(dest)
    list =glob.glob(os.path.join(src,'*' ))
    for dir in list:
        if os.path.isdir(dir):
            outpath = os.path.join(dest,os.path.basename(dir))

            list = findFiles(dir)
            if hashCheck(list, outpath+'.pyxd', 0):
                print("build asset database {}.pyxd".format(outpath))
                compressFolder(dir, outpath)
            else:
                print("skip asset database {}.pyxd".format(outpath))
        else:
            print("copy {}".format(dir))
            shutil.copy2(dir, dest)


def findConvertReplaceTextures(fig : core.editableFigure, sourceRoot, destRoot, platform=core.TARGET_PLATFORM_PC, figurePath=""):
    """
    Find and convert the texture file linked from editableFigure.

    :param sourceRoot: Root folder to search for texture files
    :param destRoot: Root folder to save the textures
    :param editableFigure: The location where the editableFigure was saved (optional). It must be included under sourceRoot.
    :param platform: Formats to convert textures to.　TARGET_PLATFORM_PC or TARGET_PLATFORM_MOBILE

    :return: None

    The figurePath must always be under the sourceRoot.
    The specified folder hierarchy may contain multiple texture files with the same name.
    When figurePath is specified, the texture file closest to it will be selected.

    Example.
        ・File path saved in editableFigure
            D:/aaa/bbb/ccc/test.png
        ・Location of the actual file
            C:/user/input/images/test.png
        ・sourceRoot
            C:/user/input
        ・destRoot
            C:/user/output
        ・The actual output file
            C:/user/output/images/test.pyxi
        ・New file path to be saved in editableFigure
            images/test
    """
    sourceRoot = os.path.abspath(sourceRoot).replace('\\', '/')
    destRoot = os.path.abspath(destRoot).replace('\\', '/')

    if figurePath == "":
        daepath = figurePath = sourceRoot
    else:
        figurePath = os.path.abspath(figurePath).replace('\\', '/')
        daepath, _ = os.path.split(figurePath)
        if len(daepath) == 0 : daepath = os.path.abspath(".").replace('\\', '/')


    imageList = findImageFiles(sourceRoot)
    imageDirectoryTree = createDirectoryTree(imageList, sourceRoot)
    setMaterialAlpha(fig, imageDirectoryTree, daepath, sourceRoot)

    allTextures = dict()
    replaceTexturePath(fig, imageDirectoryTree, daepath, sourceRoot, allTextures, removeRoot(destRoot,sourceRoot))
    convertImages(allTextures, sourceRoot, destRoot, platform)


def convertAssets(sourceDir, destDir, platform, unit=1.0, rootDir=None):
    """
    Convert model and image data to ige format
    model(.dae) -> .pyxf
    image(.png .dds .tga .psd ) -> .pyxi

    If model data with the same name as the folder name is found,
    it is regarded as base model data, and other model data in the
    same folder are regarded as motion files, and all are packed
    into one file
    If there is no model data with the same name as the folder name,
    Each model data as base model data and convert separately

    :param sourceDir: source root folder
    :param destDir: destination root folder
    :param platform: target platform (core.TARGET_PLATFORM_XX)
    :param unit: scale unit size of 3d model
    :return: None
    """
    if rootDir is None: rootDir = destDir

    imageList = findImageFiles(sourceDir)
    sceneList = find3DSceneFiles(sourceDir)
    imageDirectoryTree = createDirectoryTree(imageList, sourceDir)
    confTree = createFigureConfTree(sourceDir, unit)
    allTextures = dict()
    tmpfig = core.editableFigure("tmpfigure", True)

    print('----------------------')
    print('Convert 3d scene files')
    print('----------------------')

    for files in sceneList.values():
        tmpfig.clear()
        conf = findFigureConf(files[0], confTree)
        load3DScene(tmpfig, files, conf)

        daepath, _ = os.path.split(files[0])
        setMaterialAlpha(tmpfig, imageDirectoryTree, daepath, sourceDir)
        replaceTexturePath(tmpfig, imageDirectoryTree, daepath, sourceDir, allTextures, removeRoot(destDir,rootDir))

        outfile = os.path.normpath(files[0].replace(sourceDir, destDir, 1))
        outfile = replaceExt(outfile, '.pyxf')

        if hashCheck(files, outfile, platform):
            print("convert {}".format(files[0]))
            apputil.makeDirectories(outfile)

            if 'OPTIMIZE_VERTEX' in conf.keys() and conf['OPTIMIZE_VERTEX']:
                tmpfig.removeUnreferencedVertices()
            if 'OPTIMIZE_MESH' in conf.keys() and conf['OPTIMIZE_MESH']:
                tmpfig.mergeSameMaterialMesh()

            tmpfig.saveFigure(outfile)
        else:
            print("skip {}".format(files[0]))

    for img in imageList:
        if img not in allTextures:
            allTextures[img] = {'path': removeRoot(img,sourceDir), 'normal': False, 'wrap': False}
    convertImages(allTextures, sourceDir, destDir, platform)

    return allTextures


def deploy(projectFolder, userID, appName, appVersion, unit = 1.0, assetSource=None, assetDestination=None):
    if assetSource == None: assetSource = projectFolder
    if assetDestination == None: assetDestination = projectFolder
    allTextures = convertAssets(assetSource, assetDestination, core.TARGET_PLATFORM_PC, unit, projectFolder)

    deployTmp = '.tmp/deploy/'+appName+'/'+appVersion+'/'

    print('----------------------')
    print('Create non pack app image')
    print('----------------------')
    print('--- PC ---')
    copyFiles(projectFolder, deployTmp+'unpackimage/pc', ['.pyxf', '.json', '.pyxa', '.pyxb', '.pickle', '.zip', '.txt', '.db', '.ttf'])
    compilePrograms(projectFolder, deployTmp+'unpackimage/pc')
    convertImages(allTextures, assetSource, os.path.join(deployTmp+'unpackimage/pc',removeRoot(assetDestination, projectFolder)), core.TARGET_PLATFORM_PC)
    print('--- Mobile ---')
    copyFiles(projectFolder, deployTmp+'unpackimage/mobile', ['.pyxf', '.json', '.pyxa', '.pyxb', '.pickle', '.zip', '.txt', '.db', '.ttf'])
    compilePrograms(projectFolder, deployTmp+'unpackimage/mobile')
    convertImages(allTextures, assetSource, os.path.join(deployTmp+'unpackimage/mobile',removeRoot(assetDestination, projectFolder)), core.TARGET_PLATFORM_MOBILE)

    print('----------------------')
    print('Create packed app image')
    print('----------------------')
    print('--- PC ---')
    buildAppPackage(deployTmp+'unpackimage/pc', deployTmp+'packimage/pc')
    print('--- Mobile ---')
    buildAppPackage(deployTmp+'unpackimage/mobile', deployTmp+'packimage/mobile')

    print('----------------------')
    print('Upload app image to the server')
    print('----------------------')
    print('--- PC ---')
    launch_server.makeProjectDir(userID, appName, appVersion, "pc")
    launch_server.upload_files(deployTmp+'packimage/pc', '/'+userID+'/'+appName+'/'+appVersion+'/pc')
    print('--- Mobile ---')
    launch_server.makeProjectDir(userID, appName, appVersion, "mobile")
    launch_server.upload_files(deployTmp+'packimage/mobile',  '/'+userID+'/'+appName+'/'+appVersion+'/mobile')



"""
indi game engine \n
apprication utilities
"""
import igeCore as core
import igeVmath as vmath
import os.path


def makeDirectories(dir):
    """
    Make local directory recursive
    :param dir: file path to make directory
    :return: None
    """
    path = os.path.normpath(dir)

    list = []
    while os.path.exists(path)==False:
        if path == '':
            break
        path,tail = os.path.split(path)
        if tail =='' and tail.find('.') != -1:
            list.append(path)
            break
        else:
            list.append(tail)

    i =len(list)-1
    while i >= 0:
        name, ext = os.path.splitext(list[i])
        if ext != '':
            path = os.path.join(path, list[i])
            path = os.path.normpath(path)
            break
        else:
            path = os.path.join(path, list[i])
            path = os.path.normpath(path)
            os.mkdir(path)
        i-=1


def platformName(platform):
    """
    Return the name corresponding to the platform type

    :param platform: core.TARGET_PLATFORM_XX
    :return: string
    """
    if platform == core.TARGET_PLATFORM_PC:
        return 'pc'
    if platform == core.TARGET_PLATFORM_IOS:
        return 'ios'
    if platform == core.TARGET_PLATFORM_ANDROID:
        return 'android'
    return 'unknown'

def targetFolderName(platform):
    if platform == core.TARGET_PLATFORM_PC:
        return 'pc'
    return 'mobile'



"""
indi game engine - input module
"""
import igeCore as core
from .event import EventType
from .keyboard import KeyEventCode, KeyCode, KeyModifier, Keyboard
from .touch import TouchEventCode, Touch

def getKeyboard():
    return Keyboard.instance()

def getTouch():
    return Touch.instance()

eventListenerFuncs = set([])

def onEvent(event):
    for func in eventListenerFuncs:
        func(event)

def registerEventListener(func):
    eventListenerFuncs.add(func)
    if len(eventListenerFuncs) == 1:
        core.registerEventListener(EventType.ALL, -1, onEvent)



import importlib

try:
    importlib.import_module('numpy')
except ImportError:
    from pip._internal import main as _main
    _main(['install', 'numpy'])

try:
    importlib.import_module('conans')
except ImportError:
    from pip._internal import main as _main
    _main(['install', 'conan'])

from setuptools import setup, Extension, find_packages
from setuptools.command.build_ext import build_ext as build_ext_orig
import pathlib
import setuptools
import numpy
import sys
import os
import shutil
import platform

from os import path
here = path.abspath(path.dirname(__file__))

from conanfile import IgeConan

def setEnv(name, value):
    if platform.system() == 'Windows':
        os.system(f'set {name}={value}')
    else:
        os.system(f'export {name}={value}')
    os.environ[str(name)] = str(value)

class CMakeExtension(Extension):
    def __init__(self, name):
        super().__init__(name, sources=[])

class build_ext(build_ext_orig):
    def run(self):
        for ext in self.extensions:
            self.build_cmake(ext)
        super().run()

    def build_cmake(self, ext):
        setEnv('CONAN_REVISIONS_ENABLED', '1')
        cwd = pathlib.Path().absolute()

        # these dirs will be created in build_py, so if you don't have
        # any python sources to bundle, the dirs will be missing
        extdir = os.path.abspath(os.path.dirname(self.get_ext_fullpath(ext.name)))
        build_temp =  os.path.abspath(os.path.dirname(self.build_temp))

        # example of cmake args
        config = 'Debug' if self.debug else 'Release'
        cmake_args = [
            '-DCMAKE_LIBRARY_OUTPUT_DIRECTORY=' + str(extdir),
            '-DCMAKE_BUILD_TYPE=' + config,
            '-DAPP_STYLE=SHARED',
            '-DPYTHON_VERSION=' + str(sys.version_info[0]) + '.' + str(sys.version_info[1])
        ]

        # example of build args
        build_args = [
            '--config', config
        ]

        if not os.path.exists(self.build_temp):
            os.makedirs(self.build_temp)
        os.chdir(str(build_temp))

        conanProfile = None
        cmake_arch = None

        if sys.platform == 'win32':
            if sys.maxsize > 2 ** 32:
                conanProfile = os.path.join(str(cwd), 'cmake', 'profiles', 'windows_x86_64')
                cmake_arch = 'x64'
            else:
                conanProfile = os.path.join(str(cwd), 'cmake', 'profiles', 'windows_x86')
                cmake_arch = 'Win32'

        if conanProfile:
            self.spawn(['conan', 'install', f'--profile={conanProfile}', str(cwd), '--update', '--remote', 'ige-center'])
        else:
            self.spawn(['conan', 'install', '--update', '--remote', 'ige-center'])

        if cmake_arch:
            self.spawn(['cmake', '-A', cmake_arch, str(cwd)] + cmake_args)
        else:
            self.spawn(['cmake', str(cwd)] + cmake_args)

        ext_name = str(ext.name).split('.')[-1]
        if not self.dry_run:
            self.spawn(['cmake', '--build', '.', '--target', ext_name] + build_args)
        pyd_path = os.path.join(build_temp, config, f'{ext_name}.pyd')
        extension_path = os.path.join(cwd, self.get_ext_fullpath(ext.name))
        extension_dir = os.path.dirname(extension_path)
        if not os.path.exists(extension_dir):
            os.makedirs(extension_dir)
        shutil.move(pyd_path, extension_path)
        
        # Copy additional dlls
        for root, dirs, files in os.walk(os.path.join(build_temp, config)):
            for file in files:
                if file.endswith(".dll"):
                    if not os.path.exists(os.path.join(extension_dir, os.path.basename(file))):
                        shutil.copy(os.path.join(root, file), extension_dir)

        # Troubleshooting: if fail on line above then delete all possible
        # temporary CMake files including "CMakeCache.txt" in top level dir.
        os.chdir(str(cwd))


setup(name=IgeConan.name, version=IgeConan.version,
      description='Indi Games Engine core module',
      author=u'Indigames',
      author_email='dev@indigames.net',
      packages=find_packages(),
      ext_modules=[CMakeExtension('igeCore._igeCore'), CMakeExtension('igeCore.devtool._igeTools')],
      cmdclass={
          'build_ext': build_ext,
      },
      long_description=open('README.md').read(),
      url='https://indigames.net/',
      license='MIT',
      install_requires=['igeVmath', 'requests', 'numpy'],
      classifiers=[
          'Intended Audience :: Developers',
          'License :: OSI Approved :: MIT License',
          'Programming Language :: Python :: 3',
          #'Operating System :: MacOS :: MacOS X',
          #'Operating System :: POSIX :: Linux',
          'Operating System :: Microsoft :: Windows',
          'Topic :: Games/Entertainment',
      ],
      package_data={'igeCore': ['*.dll', 'devtool/*.dll']},
      include_package_data=True,
      setup_requires=['wheel']
      )
